/*
 * Command for accessing SPI flash.
 *
 * Copyright (C) 2008 Atmel Corporation
 * Licensed under the GPL-2 or later.
 */

#include <common.h>
#include <command.h>
#include <spi_flash.h>
#include <asm/arch/cpu.h>

static struct spi_flash *sflash;

static int do_spi_flash_probe(int argc, char *argv[])
{
	unsigned int bus 	= 0;
	unsigned int cs		= 0;
	unsigned int speed 	= CONFIG_SNOR_DEFAULT_SPEED;
	unsigned int mode  	= CONFIG_SNOR_DEFAULT_MODE;

	char *endp;
	struct spi_flash *new;

	if (argc < 2)
		goto usage;

	cs = simple_strtoul(argv[1], &endp, 0);
	if (*argv[1] == 0 || (*endp != 0 && *endp != ':'))
		goto usage;
	if (*endp == ':') {
		if (endp[1] == 0)
			goto usage;

		bus = cs;
		cs = simple_strtoul(endp + 1, &endp, 0);
		if (*endp != 0)
			goto usage;
	}

	if (argc >= 3) {
		speed = simple_strtoul(argv[2], &endp, 0);
		if (*argv[2] == 0 || *endp != 0)
			goto usage;
	}
	if (argc >= 4) {
		mode = simple_strtoul(argv[3], &endp, 16);
		if (*argv[3] == 0 || *endp != 0)
			goto usage;
	}

	new = spi_flash_probe(bus, cs, speed, mode, SPINOR);
	if (!new) {
		printf("Failed to initialize SPI flash at %u:%u\n", bus, cs);
		return 1;
	}

	if (sflash)
		spi_flash_free(sflash);
	sflash = new;

	printf("%u KiB %s at %u:%u is now current device\n",
			sflash->size >> 10, sflash->name, bus, cs);

	return 0;

usage:
	printf("Usage: snor probe [bus:]cs [hz] [mode]\n");
	return 1;
}

static int do_spi_flash_read_write(int argc, char *argv[])
{
	unsigned long offset;
	unsigned long len;
	void *buf;
	char *endp;
	int ret;

	if (argc < 3)
		goto usage;

	offset = simple_strtoul(argv[1], &endp, 16);
	if (*argv[1] == 0 || *endp != 0)
		goto usage;
	len = simple_strtoul(argv[2], &endp, 16);
	if (*argv[2] == 0 || *endp != 0)
		goto usage;

	buf = (unsigned char *) STAGE2_LOADADDR;

	if (strcmp(argv[0], "read") == 0)
		ret = spi_flash_read(sflash, offset, len, buf);
	else
		ret = spi_flash_write(sflash, offset, len, buf);

	if (ret) {
		printf("SPI flash %s failed\n", argv[0]);
		return 1;
	}

	return 0;

usage:
	printf("Usage: snor %s addr offset len\n", argv[0]);
	return 1;
}

static int do_spi_flash_erase(int argc, char *argv[])
{
	unsigned long offset;
	unsigned long len;
	char *endp;
	int ret;

	if (argc < 3)
		goto usage;

	offset = simple_strtoul(argv[1], &endp, 16);
	if (*argv[1] == 0 || *endp != 0)
		goto usage;
	len = simple_strtoul(argv[2], &endp, 16);
	if (*argv[2] == 0 || *endp != 0)
		goto usage;

	ret = spi_flash_erase(sflash, offset, len);
	if (ret) {
		printf("SPI flash %s failed\n", argv[0]);
		return 1;
	}

	return 0;

usage:
	printf("Usage: snor erase offset len\n");
	return 1;
}

static int do_spi_flash(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	const char *cmd;

	/* need at least two arguments 
	 */
	if (argc < 2) {
		goto usage;
	}

	cmd = argv[1];

	if (strcmp(cmd, "probe") == 0) {
		return do_spi_flash_probe(argc - 1, argv + 1);
	}

	/* The remaining commands require a selected device */
	if (!sflash) {
		printf("No SPI flash selected. Please run `snor probe'\n");
		return 1;
	}

	if (strcmp(cmd, "read") == 0 || strcmp(cmd, "write") == 0) 
		return do_spi_flash_read_write(argc - 1, argv + 1);
	if (strcmp(cmd, "erase") == 0) 
		return do_spi_flash_erase(argc - 1, argv + 1);

usage:
	cmd_usage(cmdtp);
	return 1;
}

U_BOOT_CMD(
		snor, 5,  1,  do_spi_flash,
		"SPI NOR sub-system",
		"probe [bus:]cs [hz] [mode]\n"
		"snor  read offset len\n"
		"snor  write offset len\n"
		"snor  erase offset len"
		);

