/* -------------------------------------------------------------
 * Copyright 2003-2010 Netlogic Microsystems Inc. (“Netlogic”).
 * -------------------------------------------------------------
 * This is a derived work from software originally provided by  
 * the external entity identified below. The licensing terms and  
 * warranties specified in the header of the original work apply 
 * to this derived work.
 * --------------------------#NETL_1#---------------------------
 */
/*
 * SPI flash interface
 *
 * Copyright (C) 2008 Atmel Corporation
 * Licensed under the GPL-2 or later.
 */
#include <common.h>
#include <spi_flash.h>
#include <spi_flash_internal.h>
#include <asm/arch/cpu.h>

int spi_flash_cmd(struct spi_slave *spi, unsigned char cmd, void *response, size_t len)
{
	unsigned long flags = SPI_XFER_BEGIN;
	int ret;

	if (len == 0)
		flags |= SPI_XFER_END;

	ret = spi_xfer(spi, 8, &cmd, NULL, flags);
	if (ret) {
		printf("SPI: Failed to send command %02x: %d\n", cmd, ret);
		return ret;
	}

	if (len) {
		ret = spi_xfer(spi, len * 8, NULL, response, SPI_XFER_END);
		if (ret)
			printf("SPI: Failed to read response (%zu bytes): %d\n",
					len, ret);
	}

	return ret;
}

int spi_flash_cmd_read(struct spi_slave *spi, const unsigned char *cmd,
		size_t cmd_len, void *data, size_t data_len)
{
	unsigned long flags = SPI_XFER_BEGIN;
	int ret;

	if (data_len == 0)
		flags |= SPI_XFER_END;

	ret = spi_xfer(spi, cmd_len * 8, cmd, NULL, flags);
	if (ret) {
		printf("SPI: Failed to send read command (%zu bytes): %d\n",
				cmd_len, ret);
	} else if (data_len != 0) {
		ret = spi_xfer(spi, data_len * 8, NULL, data, SPI_XFER_END);
		if (ret)
			printf("SPI: Failed to read %zu bytes of data: %d\n",
					data_len, ret);
	}

	return ret;
}

int spi_flash_cmd_write(struct spi_slave *spi, const unsigned char *cmd, size_t cmd_len,
		const void *data, size_t data_len)
{
	unsigned long flags = SPI_XFER_BEGIN;
	int ret;

	if (data_len == 0)
		flags |= SPI_XFER_END;

	ret = spi_xfer(spi, cmd_len * 8, cmd, NULL, flags);
	if (ret) {
		printf("SPI: Failed to send read command (%zu bytes): %d\n",
				cmd_len, ret);
	} else if (data_len != 0) {
		ret = spi_xfer(spi, data_len * 8, data, NULL, SPI_XFER_END);
		if (ret)
			printf("SPI: Failed to read %zu bytes of data: %d\n",
					data_len, ret);
	}

	return ret;
}

int spi_flash_read_common(struct spi_flash *flash, const unsigned char *cmd,
		size_t cmd_len, void *data, size_t data_len)
{
	struct spi_slave *spi = flash->spi;
	int ret;

	spi_claim_bus(spi);
	ret = spi_flash_cmd_read(spi, cmd, cmd_len, data, data_len);
	spi_release_bus(spi);

	return ret;
}

/* spi_flash_probe is used by both
 * SPI NAND and SPI NOR devices on XLP
 */
struct spi_flash *spi_flash_probe(unsigned int bus, unsigned int cs,
		unsigned int max_hz, unsigned int spi_mode, unsigned int device)
{
	struct spi_slave *spi;
	struct spi_flash *flash;
	int ret;
	unsigned char idcode[5];

	/* common slave init routine for
	 * SPI devices from drivers/spi.c
	 */
	spi = spi_setup_slave(bus, cs, max_hz, spi_mode, device);
	if (!spi) {
		printf("SPI: Failed to set up slave\n");
		return NULL;
	}

	if (device == SPINAND) {
		spi_flash_reset(spi);
	}

	ret = spi_claim_bus(spi);
	if (ret) {
		printf("SPI: Failed to claim SPI bus: %d\n", ret);
		goto err_claim_bus;
	}

	/* Read the ID codes for SPI NAND
	 * or SPI NOR, based on how we are
	 * called from cmdline.
	 */
	if (device == SPINAND) {
		ret = spi_flash_cmd(spi, CMD_READ_ID, 
				&idcode, 3);
		idcode[0] = idcode[1];
		idcode[1] = idcode[2];
	}
	else {
		ret = spi_flash_cmd(spi, CMD_READ_ID, 
				&idcode, sizeof(idcode));
	}
	if (ret)
		goto err_read_id;

	if (device == SPINAND) {
		printf("SPI: Got idcode %02x %02x\n", 
				idcode[0], idcode[1]);
	} else {
		printf("SPI: Got idcode %02x %02x %02x %02x %02x\n", 
				idcode[0], idcode[1], idcode[2], idcode[3], idcode[4]);
	}

	switch (idcode[0]) {
		case 0x01:
			/* Found Spansion NOR part */
			flash = spi_flash_probe_spansion(spi, idcode);
			break;
		case 0x2C:
			/* Found Micron NAND part */
			flash = spi_flash_probe_micron(spi, idcode);
			break;
		default:
			printf("SPI: Unsupported manufacturer %02X\n", idcode[0]);
			flash = NULL;
			break;
	}

	if (!flash)
		goto err_manufacturer_probe;

	spi_release_bus(spi);

	return flash;

err_manufacturer_probe:
err_read_id:
	spi_release_bus(spi);
err_claim_bus:
	spi_free_slave(spi);
	return NULL;
}

void spi_flash_free(struct spi_flash *flash)
{
	spi_free_slave(flash->spi);
}
