/*
 * ADD LICENSING HEADER HERE...
 */
#include <common.h>
#include <spi_flash.h>
#include "spi_flash_internal.h"

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

#define CMD_MT29F_RD_ARRAY    	0x03    /* Read Data Bytes */
#define CMD_MT29F_WREN        	0x06    /* Write Enable */
#define CMD_MT29F_WRDI        	0x04    /* Write Disable */
#define CMD_MT29F_RDSR        	0x0f    /* Read Status Register */
#define CMD_MT29F_WRSR        	0x1f    /* Write Status Register */
#define CMD_MT29F_PP          	0x02    /* Page Program */
#define CMD_MT29F_PE          	0x10    /* Page Execute */
#define CMD_MT29F_BE          	0xd8    /* Block Erase */
#define CMD_MT29F_RESET       	0xff    /* Issue Reset */

#define CMD_READ_ARRAY        	0x03
#define CMD_PAGE_READ         	0x13

#define MICRON_ID_MT29F1G01ZAC	0x12
#define MICRON_SR_WIP         	(1<<0)  /* Write-in-Progress */
#define SPINAND_TIMEOUT			0x7fffffff

struct micron_spi_flash_params {
	u8 	idcode1;
	u16 page_size;
	u16 pages_per_block;
	u16 nr_blocks;
	const char *name;
};

struct micron_spi_flash {
	struct spi_flash flash;
	const struct micron_spi_flash_params *params;
};
struct micron_spi_flash micron_global;

static inline struct micron_spi_flash *to_micron_spi_flash(struct spi_flash
							     *flash)
{
	return container_of(flash, struct micron_spi_flash, flash);
}

static const struct micron_spi_flash_params micron_spi_flash_table[] = {
	{
		.idcode1 		 = MICRON_ID_MT29F1G01ZAC,
		.page_size 		 = 2048,
		.pages_per_block = 64,
		.nr_blocks 		 = 1024,
		.name 			 = "MT29F1G01ZAC",
	},
};

int spi_read_status(struct spi_slave *spi, void *response)
{
	unsigned long flags = SPI_XFER_BEGIN;
	int ret;
	unsigned char cmd[2];

	cmd[0] = CMD_MT29F_RDSR; 
	cmd[1] = 0xc0; 

	ret = spi_xfer(spi, 16, &cmd, NULL, flags);
	if (ret) {
		printf("SF: Failed to send command %02x: %d\n", cmd[0], ret);
		return ret;
	}

	ret = spi_xfer(spi,  8, NULL, response, SPI_XFER_END);
	if (ret)
		printf("SF: Failed to read response (8 bytes): %d\n",ret);

	return ret;
}

static int micron_wait_ready(struct spi_slave *spi, unsigned long timeout)
{
	int ret;
	unsigned char status;

	do {
		ret = spi_read_status(spi, &status);
		if (ret)
			return -1;

		if ((status & MICRON_SR_WIP) == 0)
			break;
		timeout-- ;

	} while (timeout != 0);

	if ((status & MICRON_SR_WIP) == 0)
		return 0;

	/* Timed out */
	return -1;
}

/* TODO: Make this a device-specific
 * reset routine
 */
int spi_flash_reset(struct spi_slave *spi)
{
	int ret, i;

	ret = spi_flash_cmd(spi, CMD_MT29F_RESET, NULL, 0);
	if (ret < 0) {
		printf("SF: Reset failed\n");
	}
	
	/* Wait 1ms */
	for (i=0; i<2000000; i++) {};
	
	ret = micron_wait_ready(spi, SPINAND_TIMEOUT);
	if (ret < 0) {
		printf("SPI NAND: Timeout\n");
		return ret ;
	}

	return ret ;
}

static int micron_read(struct spi_flash *flash,
			     u32 offset, size_t len, void *buf)
{
	struct micron_spi_flash *micron = to_micron_spi_flash(flash);

	unsigned long page_addr;
	unsigned long page_size;
	unsigned long byte_addr;
	unsigned int  block_size;
	size_t chunk_len, actual;
	int ret = 0;
	
	unsigned char cmd[4];

	page_size 		= micron->params->page_size;
	block_size 		= page_size * (micron->params->pages_per_block);

	/* extract page/offset
	 * we need to read from
	 */
	page_addr 		= offset / page_size;
	byte_addr 		= offset % page_size;

	for (actual = 0; actual < len; actual += chunk_len) {

		chunk_len = min(len - actual, page_size - byte_addr);

		/* Set the page we want to read from...  
		 */
		cmd[0] = CMD_PAGE_READ;
		cmd[1] = 0x00;				/* dummy bits */
		cmd[2] = page_addr >> 8;
		cmd[3] = page_addr;

		/* Send this command across 
		 */
		ret = spi_flash_cmd_write(flash->spi, cmd, 4, NULL, 0) ;
		if (ret < 0) {
			printf("SPI NAND: Page Read failed\n");
			return ret ;
		}

		/* Check for Status 
		 */
		ret = micron_wait_ready(flash->spi, SPINAND_TIMEOUT);
		if (ret < 0) {
			printf("SPI NAND: MICRON page read timed out\n");
			return ret ;
		}

		cmd[0] = CMD_READ_ARRAY;
		cmd[1] = (byte_addr % page_size) >> 8;
		cmd[2] = (byte_addr % page_size);
		cmd[3] = 0x0;
#if 0
		printf("READ: 0x%x => cmd = { 0x%02x 0x%02x%02x%02x } len = 0x%x\n",
			 offset, cmd[0], cmd[1], cmd[2], cmd[3], len);
#endif
		ret |= spi_flash_read_common(flash, cmd, 4, buf + actual, chunk_len);

		page_addr++;
		byte_addr = 0;
	}
	
	return ret;
}

static int micron_write(struct spi_flash *flash,
			 u32 offset, size_t len, const void *buf)
{
#if 0
	struct micron_spi_flash *micron = to_micron_spi_flash(flash);
	unsigned long page_addr;
	unsigned long byte_addr;
	unsigned long page_size;
	size_t chunk_len;
	size_t actual;
	int ret;
	unsigned char cmd[4];

	page_size = micron->params->page_size;
	page_addr = offset / page_size;
	byte_addr = offset % page_size;

	ret = spi_claim_bus(flash->spi);
	if (ret) {
		printf("SF: Unable to claim SPI bus\n");
		return ret;
	}

	ret = 0;
	for (actual = 0; actual < len; actual += chunk_len) {
		chunk_len = min(len - actual, page_size - byte_addr);

		cmd[0] = CMD_S25FLXX_PP;
		cmd[1] = page_addr >> 8;
		cmd[2] = page_addr;
		cmd[3] = byte_addr;

		printf
		    ("PP: 0x%p => cmd = { 0x%02x 0x%02x%02x%02x } chunk_len = %d\n",
		     buf + actual, cmd[0], cmd[1], cmd[2], cmd[3], chunk_len);

		ret = spi_flash_cmd(flash->spi, CMD_S25FLXX_WREN, NULL, 0);
		if (ret < 0) {
			printf("SF: Enabling Write failed\n");
			break;
		}

		ret = spi_flash_cmd_write(flash->spi, cmd, 4,
					  buf + actual, chunk_len);
		if (ret < 0) {
			printf("SF: SPANSION Page Program failed\n");
			break;
		}

		ret = spansion_wait_ready(flash, SPINAND_TIMEOUT);
		if (ret < 0) {
			printf("SF: SPANSION page programming timed out\n");
			break;
		}

		page_addr++;
		byte_addr = 0;
	}

	printf("SF: SPANSION: Successfully programmed %u bytes @ 0x%x\n",
	      len, offset);

	spi_release_bus(flash->spi);
	return ret;
#else
	return 0;
#endif
}

int micron_erase(struct spi_flash *flash, u32 offset, size_t len)
{
#if 0
	struct micron_spi_flash *micron = to_micron_spi_flash(flash);
	unsigned long sector_size;
	size_t actual;
	int ret;
	unsigned char cmd[4];

	/*
	 * This function currently uses sector erase only.
	 * probably speed things up by using bulk erase
	 * when possible.
	 */

	sector_size = micron->params->page_size * micron->params->pages_per_block;

	if (offset % sector_size || len % sector_size) {
		printf("SF: Erase offset/length not multiple of sector size\n");
		return -1;
	}

	len /= sector_size;
	cmd[0] = CMD_S25FLXX_SE;
	cmd[2] = 0x00;
	cmd[3] = 0x00;

	ret = spi_claim_bus(flash->spi);
	if (ret) {
		printf("SF: Unable to claim SPI bus\n");
		return ret;
	}

	ret = 0;
	for (actual = 0; actual < len; actual++) {
		cmd[1] = (offset / sector_size) + actual;

		ret = spi_flash_cmd(flash->spi, CMD_S25FLXX_WREN, NULL, 0);
		if (ret < 0) {
			printf("SF: Enabling Write failed\n");
			break;
		}

		ret = spi_flash_cmd_write(flash->spi, cmd, 4, NULL, 0);
		if (ret < 0) {
			printf("SF: SPANSION page erase failed\n");
			break;
		}

		/* Up to 2 seconds */
		ret = spansion_wait_ready(flash, SPI_FLASH_PAGE_ERASE_TIMEOUT);
		if (ret < 0) {
			printf("SF: SPANSION page erase timed out\n");
			break;
		}
	}

	printf("SF: SPANSION: Successfully erased %lu bytes @ 0x%x\n",
	      len * sector_size, offset);

	spi_release_bus(flash->spi);
	return ret;
#else
	return 0;
#endif	
}
	
struct spi_flash *spi_flash_probe_micron(struct spi_slave *spi, unsigned char *idcode)
{
	const struct micron_spi_flash_params *params;
	u32 i;
	u8 device_id = idcode[1];

	/* Is this part supported..?
	 */
	for (i = 0; i < ARRAY_SIZE(micron_spi_flash_table); i++) {
		params = &micron_spi_flash_table[i];
		if (params->idcode1 == device_id) {
				break;
		}
	}

	if (i == ARRAY_SIZE(micron_spi_flash_table)) {
		printf("SF: Unsupported MICRON ID %04x\n", device_id);
		return NULL;
	}

	micron_global.params 		= params;
	micron_global.flash.spi 	= spi;
	micron_global.flash.name 	= params->name;

	micron_global.flash.write 	= micron_write;
	micron_global.flash.erase 	= micron_erase;
	micron_global.flash.read 	= micron_read;
	micron_global.flash.size 	= (params->page_size * params->pages_per_block
	    * params->nr_blocks);

	printf("SPI NAND: Detected %s with page size %u, total %u bytes\n",
	      params->name, params->page_size, micron_global.flash.size);

	return &micron_global.flash;
}
