/*
 * Copyright 2003-2010 Netlogic Microsystems (“Netlogic”). All rights
 * reserved. Redistribution and use in source and binary forms, with 
 * or without modification, are permitted provided that the following 
 * conditions are met:
 *
 *	1. 	Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 *
 *	2. 	Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in
 *		the documentation and/or other materials provided with the
 *		distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * --------------------------------#NETL_2#--------------------------------
 */

/* This file has the DRAM programming with
 * added hooks for supporting multiple nodes
 * and channels across dual I2C busses. 
 */

#include <common.h>
#include <asm/arch/cpu.h>

#include "ddr_spd.h"
#include "clock.h"
#include "ddr_helper.h"
#include "ddr_spd.h"
#include "ddr_leveling.h"

#define NUM_SPD_BYTES   	256
#define DDR_MAX_PERIOD_PS       1500

struct spd_data {
	    __u8 spd_byte[NUM_SPD_BYTES];
};
struct spd_data spd_dimm0;

struct spd_mapping {
	int32_t bus;
	int32_t spd_addr;
};

static const struct spd_mapping 
	dimm_to_spd_addr[MAX_NODES][DDR_NUM_CHANNELS][DDR_NUM_DIMMS] = 
{ { { {0,0x52}, {0,0x56} },      /* Node 0:  Channel A (0) DIMM0, DIMM1 */
    { {1,0x52}, {1,0x56} },      /* Node 0:  Channel B (1) DIMM0, DIMM1 */
    { {0,0x51}, {0,0x55} },      /* Node 0:  Channel C (2) DIMM0, DIMM1 */
    { {1,0x51}, {1,0x55} } },    /* Node 0:  Channel D (3) DIMM0, DIMM1 */
 {  { {2,0x52}, {2,0x56} },      /* Node 1:  Channel A (0) DIMM0, DIMM1 */
    { {3,0x52}, {3,0x56} },      /* Node 1:  Channel B (1) DIMM0, DIMM1 */
    { {2,0x51}, {2,0x55} },      /* Node 1:  Channel C (2) DIMM0, DIMM1 */
    { {3,0x51}, {3,0x55} } },    /* Node 1:  Channel D (3) DIMM0, DIMM1 */
 {  { {4,0x52}, {4,0x56} },      /* Node 1:  Channel A (0) DIMM0, DIMM1 */
    { {5,0x52}, {5,0x56} },      /* Node 1:  Channel B (1) DIMM0, DIMM1 */
    { {4,0x51}, {4,0x55} },      /* Node 1:  Channel C (2) DIMM0, DIMM1 */
    { {5,0x51}, {5,0x55} } },    /* Node 1:  Channel D (3) DIMM0, DIMM1 */
 {  { {6,0x52}, {6,0x56} },      /* Node 1:  Channel A (0) DIMM0, DIMM1 */
    { {7,0x52}, {7,0x56} },      /* Node 1:  Channel B (1) DIMM0, DIMM1 */
    { {6,0x51}, {6,0x55} },      /* Node 1:  Channel C (2) DIMM0, DIMM1 */
    { {7,0x51}, {7,0x55} } }     /* Node 1:  Channel D (3) DIMM0, DIMM1 */
} ;

/*
 * SPD read
 *
 */
static int read_ddr_spd(int    bus,
			 __u8   eeprom_addr, 
			 __u16  eeprom_offset,
			 int    numbytes, 
			 __u8   *buf) {
	
	return (oc_i2c_rx(bus,eeprom_addr, eeprom_offset, 
			  numbytes, buf));
}

/*
 * SPD CRC16
 *
 */
static int spd_crc16(u8 *ptr, 
		     int count)
{
	int crc=0, i;

	while (--count >= 0) {
		crc = crc ^ (int)*ptr++ << 8;
		for (i=0; i<8; ++i)
			if (crc & 0x8000)
				crc = crc << 1 ^ 0x1021;
			else
				crc = crc << 1;
	}
	return (crc & 0xFFFF);
}

/*
 * Verify if SPD is a legal DDR3 type and CRC is clean
 *
 */
static int verify_spd(u8* spd)
{
	int crc_bytes  = 0;
	int crc_cal    = 0;
	int crc_expect = 0;

	if (!(SPD_DDR3(spd))) {
		printf ("Error! DIMM not DDR3 type\n");
		return -1;
	}

	crc_bytes  = SPD_CRC_117(spd) ? 117:126;
	crc_cal    = spd_crc16(spd,crc_bytes);
	crc_expect = SPD_CRC(spd);

	if (crc_cal !=  crc_expect) {
		printf("Error! SPD crc16 check failed. Calc=%x, Exp=%x\n",
				crc_cal,crc_expect);
		return -1;
	}
	return 0;
}

/*
 * Verify compatible DDR3 parameters
 *
 */
static int merge_param(int  init, 
		       u32* old_param, 
		       u32  new_param) 
{
	if (!init) {
		if (*old_param != new_param) {
			printf("Error! Mixing incompatible DIMM types\n"); 
			return -1;
		}
	} else {
		*old_param = new_param;
	}
	return 0;
}

/*
 * Convert SPD timebased values to clock cycles
 *
 */
static u32 convert_cycles(u32 val, 
			  u32 timebase_ps, 
			  u32 clock_period_ps, 
			  u8 round_up) 
{	
	u32 time_ps = (val * timebase_ps);

	if (((time_ps%clock_period_ps) == 0) ||
			(round_up == 0)) {
		return (time_ps/clock_period_ps);
	} else {
		return ((time_ps/clock_period_ps) + 1);
	}
}

/*
 * Load DDR3 parameter structure based upon SPD values
 *
 */
static int param_load(struct ddr_params* params, 
		      const u8* spd, 
		      u8 dimm) 
{	
	int cfg_err=0, value=0;

	int initial_cfg = (params->num_ranks == 0); 

	if (initial_cfg) {
		params->rank_vec=0;
		params->mirrored=0;
	}

	switch (SPD_NUM_RANKS(spd)) {
		case 0:
			params->num_ranks += 1;
			params->rank_vec  |= (1 << (dimm * 2));
			params->quad_rank  = 0;
			break;
		case 1:
			params->num_ranks += 2;
			params->rank_vec  |= (3 << (dimm * 2));
			params->quad_rank  = 0;
			break;
		case 3:
			if (initial_cfg) {
				params->num_ranks = 4;
				params->rank_vec  = 0xf;
				params->quad_rank = 1;
			} else {
				cfg_err = 1;
			}
			break;
		default: 
			cfg_err = 1; 
			break;
	}

	if (cfg_err) {
		printf("Error: Unsupported RANK configuration %d\n",
				SPD_NUM_RANKS(spd));
		return -1;	
	}

	if ( (SPD_DIMM_TYP(spd) == SPD_RDIMM_TYPE)    ||
	     (SPD_DIMM_TYP(spd) == SPD_MINRDIMM_TYPE) ||
	     (SPD_DIMM_TYP(spd) == SPD_72BSORDIMM_TYPE) ) {
		value = 1;
	} else {
		value = 0;
	}

	if (merge_param(initial_cfg,
		&params->registered, value)) { 
		return -1; 
	};

	if (!params->registered) {
		if (SPD_MIRROR(spd) == 1) {
			if (params->num_ranks == 2) {
				params->mirrored = 0x2;
			}  else {
				params->mirrored = 0xa;
			}
		} 
	}

	switch (SPD_DEV_WIDTH(spd)) {

		case SPD_DEV_X8 : 
			value = 8; break;
		case SPD_DEV_X16: 
			value = 16; break;
		case SPD_DEV_X32: 
			value = 32; break;
		default:
 			printf("Error: Unsupported device width %d\n",
					SPD_DEV_WIDTH(spd));
 			return -1;
	}

	if (merge_param(initial_cfg,
		&params->dev_width,value)) { 
		return -1; 
	};

	switch (SPD_MOD_WIDTH(spd)) {
		case SPD_MOD_X32: 
			value = 32; break; 
		case SPD_MOD_X64: 
			value = 64; break; 
		default:
		 	printf("Error! Unsupported module width %d\n",
					SPD_DEV_WIDTH(spd));
		  	return -1;
	}

	if (merge_param(initial_cfg,
		&params->mod_width, value)) { 
		return -1; 
	};

	params->ecc_dev    = initial_cfg ? SPD_ECC(spd):(params->ecc_dev && SPD_ECC(spd));
	params->vdd135_cap = initial_cfg ? SPD_135V(spd):(params->vdd135_cap && SPD_135V(spd));
	params->rzq_div6   = initial_cfg ? SPD_RZQD6(spd):(params->rzq_div6 && SPD_RZQD6(spd));
	params->rzq_div7   = initial_cfg ? SPD_RZQD7(spd):(params->rzq_div7 && SPD_RZQD7(spd));

	if (merge_param(initial_cfg,
		&params->node_density_mb, 
		(256 << SPD_CAPACITY(spd)))) {
		return -1;  
	}
	
	if (merge_param(initial_cfg,
		&params->row_bits,
		(12 + SPD_ROW_BITS(spd)))) {
		return -1;
	}
	
	if (merge_param(initial_cfg,
		&params->col_bits,
		(9  + SPD_COL_BITS(spd)))) {
		return -1;
	}
	
	if (merge_param(initial_cfg,
		&params->ftb_fs,
		SPD_FTB(spd))) {
		return -1;
	}
	
	if (merge_param(initial_cfg,
		&params->mtb_ps,
		SPD_MTB(spd))) {
		return -1;
	}


	if (initial_cfg) {
		params->tck_min   = SPD_tCKmin(spd);
		params->taa_min   = SPD_tAAmin(spd);
		params->cas_vec   = SPD_CASLAT(spd);
		params->twr_min   = SPD_tWRmin(spd);
		params->trcd_min  = SPD_tRCDmin(spd);
		params->trrd_min  = SPD_tRRDmin(spd);
		params->trp_min   = SPD_tRPmin(spd);
		params->twtr_min  = SPD_tWTRmin(spd);
		params->trtp_min  = SPD_tRTPmin(spd);
		params->tfaw_min  = SPD_tFAW(spd);
		params->tras_min  = SPD_tRASmin(spd);
		params->trc_min   = SPD_tRCmin(spd);
		params->trfc_min  = SPD_tRFCmin(spd);
	} else {
		/* Take lowest frequency (highest period) */
		if (params->tck_min<SPD_tCKmin(spd))
			params->tck_min = SPD_tCKmin(spd);

		/* Take highest CAS latency */
		if (params->taa_min<SPD_tAAmin(spd))
			params->taa_min = SPD_tAAmin(spd);

		/* Take common supported CAS cas latencies */
		params->cas_vec = params->cas_vec & SPD_CASLAT(spd);

		/* Take highest write recovery time */
		if (params->twr_min<SPD_tWRmin(spd)) 
			params->twr_min = SPD_tWRmin(spd);

		/* Take highest tRCD (ras to cas delay) time */
		if (params->trcd_min<SPD_tRCDmin(spd))
			params->trcd_min = SPD_tRCDmin(spd);

		/* Take highest tRRD (active row to active row delay) time */
		if (params->trrd_min<SPD_tRRDmin(spd)) 
			params->trrd_min = SPD_tRRDmin(spd);

		/* Take highest tRP (row precharge) time */
		if (params->trp_min<SPD_tRPmin(spd)) 
			params->trp_min = SPD_tRPmin(spd);

		/* Take highest tWTR (write to read delay) time */
		if (params->twtr_min<SPD_tWTRmin(spd)) 
			params->twtr_min = SPD_tWTRmin(spd);

		/* Take highest tRTP (read to precharge delay) time */
		if (params->trtp_min<SPD_tRTPmin(spd)) 
			params->trtp_min = SPD_tRTPmin(spd);

		/* Take highest tFAW (four active window delay) time */
		if (params->tfaw_min<SPD_tFAW(spd)) 
			params->tfaw_min = SPD_tFAW(spd);

		/* Take highest tRAS (active to precharge delay) time */
		if (params->tras_min<SPD_tRASmin(spd)) 
			params->tras_min = SPD_tRASmin(spd);

		/* Take highest tRAS (active to refresh delay) time */
		if (params->trc_min<SPD_tRCmin(spd)) 
			params->trc_min = SPD_tRCmin(spd);

		/* Take highest tRFC (refresh recovery delay) time */
		if (params->trfc_min<SPD_tRFCmin(spd)) 
			params->trfc_min = SPD_tRFCmin(spd);
	}

	if (params->tck_min < (DDR_MAX_PERIOD_PS/params->mtb_ps) ) {
		params->tck_min = (DDR_MAX_PERIOD_PS/params->mtb_ps) ;
	}
	    
	/* tREF is aways 7.8us */
	params->tref_min = (7800000/params->mtb_ps);

	return 0;
}

/*
 * Configure DDR3 controller
 *
 */
static int config(struct ddr_params* params,
		  u8 node,
		  u8 chan, 
		  u32 clock)
{
	volatile u32* mmio = (volatile u32*) dmc_io_mmio(node,chan);

	u32 rank_bits=0, cas_lat=0, cas_val=0, twr_val=0;
        u32 bank_bits=0, cur_rank = 0, burst_bits=0;
	u32 val ;
	
	ddr_wr_tccd(mmio,4);

	ddr_wr_trc (mmio, convert_cycles(params->trc_min,params->mtb_ps,clock, 1));
	ddr_wr_trcd(mmio, convert_cycles(params->trcd_min,params->mtb_ps,clock,1));
	ddr_wr_tras(mmio, convert_cycles(params->tras_min,params->mtb_ps,clock,1));
	ddr_wr_trfc(mmio, convert_cycles(params->trfc_min,params->mtb_ps,clock,1));
	ddr_wr_tref(mmio, convert_cycles(params->tref_min,params->mtb_ps,clock,0));
	ddr_wr_trp (mmio, convert_cycles(params->trp_min,params->mtb_ps,clock,1));
	
	ddr_wr_tfaw(mmio,convert_cycles(params->tfaw_min,params->mtb_ps,clock,1));

	/* TXPR = trfc + 10ns
	 */
	val = (params->trfc_min + 10000/params->mtb_ps) ;
	ddr_wr_txpr(mmio,convert_cycles(val,params->mtb_ps,clock,1) ) ;
	ddr_wr_txs(mmio,convert_cycles(val,params->mtb_ps,clock,1) ) ;

	val = convert_cycles(params->trrd_min,params->mtb_ps,clock,1) ;
	
	if (val < 4) {
	  ddr_wr_trrd(mmio, 4);
	} else {
	  ddr_wr_trrd(mmio, val) ;
	}
	val = convert_cycles(params->trtp_min,params->mtb_ps,clock,1) ;
	if (val < 4) {
	  ddr_wr_trtp(mmio, 4) ;
	} else {
	  ddr_wr_trtp(mmio, convert_cycles(params->trtp_min,params->mtb_ps,clock,1));
	}
	val = convert_cycles(params->twtr_min,params->mtb_ps,clock,1) ;
	if (val < 4) {
	  ddr_wr_twtr(mmio,4) ;
	} else {
	  ddr_wr_twtr(mmio, convert_cycles(params->twtr_min,params->mtb_ps,clock,1));
	}

	
	cas_lat = convert_cycles(params->taa_min,params->mtb_ps,clock,1);

	while (cas_lat<19) {
		if (params->cas_vec & (1 << cas_lat)) {
			break;
		}
		cas_lat++;
	}
	if (cas_lat > 18) {
		printf("Error! Unable to find valid CAS setting\n");
		return -1;
	}
        switch (cas_lat) {
	case 5: cas_val = 2 ; break ;
	case 6: cas_val = 4 ; break ;
	case 7: cas_val = 6 ; break ;
	case 8: cas_val = 8 ; break ;
	case 9: cas_val = 10; break ;
	case 10: cas_val = 12; break ;
	case 11: cas_val = 14; break ;
	case 12: cas_val = 1; break ;
	case 13: cas_val = 3; break ;
	case 14: cas_val = 5; break ;
	case 15: cas_val = 7; break ;
	case 16: cas_val = 9; break ;
	default:
		printf("Error! unknown cas value %d\n",cas_lat) ; 
		break ;
	}

        ddr_wr_mr0cas(mmio,cas_val);
	
	switch (convert_cycles(params->twr_min,params->mtb_ps,clock,1) ) {
	case 5: twr_val = 1; break ;
	case 6: twr_val = 2; break ;
	case 7: twr_val = 3; break ;
	case 8: twr_val = 4; break ;
	case 10: twr_val = 5; break ;
	case 12: twr_val = 6; break ;
	case 14: twr_val = 7; break ;
	case 16: twr_val = 0; break ;
	default:
		printf("Error! unknown twr value %d\n",twr_val) ;
		break ;
	}		
	ddr_wr_mr0twr(mmio,twr_val) ;

	if ( (clock >= 750) &&
	     (clock < 833) ) {
		// tcwl = 12 (encoded 7)
		ddr_wr_mr2cwl(mmio,7);
	} else if ( (clock >= 833) &&
		    (clock < 935) ) {
		// tcwl = 11 (encoded 6)
		ddr_wr_mr2cwl(mmio,6);
	} else if ( (clock >= 935) &&
		    (clock < 1070) ) {
		// tcwl = 10 (encoded 5)
		ddr_wr_mr2cwl(mmio,5);
	} else if ( (clock >= 1070) &&
		    (clock < 1250) ) {
		// tcwl = 9 (encoded 4)
		ddr_wr_mr2cwl(mmio,4);
	} else if ( (clock >= 1250) &&
		    (clock < 1500) ) {
		// tcwl = 8 (encoded 3)
		ddr_wr_mr2cwl(mmio,3);
	} else if ( (clock >= 1500) &&
		    (clock < 1875) ) {
		// tcwl = 7 (encoded 2)
		ddr_wr_mr2cwl(mmio,2);
	} else if ( (clock >= 1875) &&
		    (clock < 2500) ) {
		// tcwl = 6 (encoded 1)
		ddr_wr_mr2cwl(mmio,1);
	} else if (clock >= 2500) { 
		// tcwl = 5 (encoded 0)
		ddr_wr_mr2cwl(mmio,0);
	} else {
		printf("Error! unknown cas write latency value: clk period %d ps\n",clock) ;
	}
	    
	bank_bits = 3;
	ddr_wr_banklsb(mmio, 0);

	if (params->mod_width == 32) {
		burst_bits = 4; 
	} else {
		burst_bits = 3 ;
	}

	if (params->num_ranks == 1) {
		rank_bits = 0;
	} else if (params->num_ranks == 2) {
		rank_bits = 1;
	} else {
		rank_bits = 2 ;
	}

	ddr_wr_ranklsb(mmio, 0 + bank_bits);
	ddr_wr_rankbits(mmio, rank_bits);

	ddr_wr_collsb(mmio, 0 + bank_bits + rank_bits);
	ddr_wr_colbits(mmio, (params->col_bits - burst_bits) );

	ddr_wr_rowlsb(mmio, 0 + bank_bits + rank_bits + (params->col_bits - burst_bits) );
	ddr_wr_rowbits(mmio, params->row_bits);

	if (params->rank_vec & 0x1) {
		ddr_wr_cs0rnkid(mmio, cur_rank);
		cur_rank++;
	} else {
		ddr_wr_cs0rnkid(mmio,3);
	}

	if (params->rank_vec & 0x2) {
		ddr_wr_cs1rnkid(mmio,cur_rank);
		cur_rank++;
	} else {
		ddr_wr_cs1rnkid(mmio,3);
	}

	if (params->rank_vec & 0x4) {
		ddr_wr_cs2rnkid(mmio, cur_rank);
		cur_rank++;
	} else {
		ddr_wr_cs2rnkid(mmio,3);
	}
	
	if (params->rank_vec & 0x8) {
		ddr_wr_cs3rnkid(mmio, cur_rank);
		cur_rank++;
	} else {
		ddr_wr_cs3rnkid(mmio,3);
	}
	
	params->rank_vec = (0xf >> (4 - params->num_ranks) ) ; 
	ddr_wr_mirror(mmio,params->mirrored);
	ddr_wr_ntaddr(mmio,3);
	ddr_wr_ntcsdly(mmio,3);
	ddr_wr_bl16(mmio,(params->mod_width == 32) ) ;
	ddr_wr_twlo(mmio,0x12) ;
        ddr_wr_twldqsen(mmio,0x1f) ;
	/* AL - 2 */
	ddr_wr_mr1al(mmio,0x2) ;

	if (params->registered) {
		/* Make Derived Parameter 8 Writable */
		ddr_rmw_reg(mmio,DDR_DP8,0x1,0x1,31) ;
		/* Override Power Down counters */
		ddr_rmw_reg(mmio,DDR_DP8,0x96282814,0xffffffff,0) ;
		
		ddr_disable_pissue(mmio) ;
	}
	
	return 0;
}

static void write_cmr(struct ddr_params* params,
		      u8 node,
		      u8 chan,
		      u32 clock)
{
	volatile u32* mmio = (volatile u32*) dmc_io_mmio(node,chan);

	/* Set CMR speed */
	if (clock == 1250) {
		ddr_write_cmr(mmio,10,DDR_CMR_SPEED1600) ;
	} else if (clock == 1500) {
		ddr_write_cmr(mmio,10,DDR_CMR_SPEED1333) ;
	} else if (clock == 1875) {
		ddr_write_cmr(mmio,10,DDR_CMR_SPEED1066) ;
	} else if (clock == 2500) { 
		ddr_write_cmr(mmio,10,DDR_CMR_SPEED800) ;
	} else if (clock == 3000) { 
		ddr_write_cmr(mmio,10,DDR_CMR_SPEED667) ;
	} else {
		printf("Error! Unknown frequency for CMR register write: clk period %d ps\n",clock) ;
	}	

	/* Add 1.35 Volt Programming */
} 
/* DDR PLL Setup:
 * Find lowest common DDR frequency   
 */	
static void find_lowest_freq(int node,
			     u32* ddr_freq_khz,
			     u32* ddr_period_ps) 
{		
	if (ddr_cntrl[0].tck_min<ddr_cntrl[1].tck_min) 
		ddr_cntrl[0].tck_min = ddr_cntrl[1].tck_min;
	
	if (ddr_cntrl[0].tck_min<ddr_cntrl[2].tck_min) 
		ddr_cntrl[0].tck_min = ddr_cntrl[2].tck_min;			
	
	if (ddr_cntrl[0].tck_min<ddr_cntrl[3].tck_min)
		ddr_cntrl[0].tck_min = ddr_cntrl[3].tck_min;
	
	ddr_cntrl[1].tck_min = ddr_cntrl[0].tck_min;
	ddr_cntrl[2].tck_min = ddr_cntrl[0].tck_min;
	ddr_cntrl[3].tck_min = ddr_cntrl[0].tck_min;
	
	if (chan_mem_size[node][0] != 0) {
		*ddr_period_ps = (ddr_cntrl[0].tck_min * ddr_cntrl[0].mtb_ps);
	} else if (chan_mem_size[node][1] != 0) {
		*ddr_period_ps = (ddr_cntrl[1].tck_min * ddr_cntrl[1].mtb_ps);
	} else if (chan_mem_size[node][2] != 0) {
		*ddr_period_ps = (ddr_cntrl[2].tck_min * ddr_cntrl[2].mtb_ps);
	} else if (chan_mem_size[node][3] != 0) {
		*ddr_period_ps = (ddr_cntrl[3].tck_min * ddr_cntrl[3].mtb_ps);
	} else {
		*ddr_period_ps = 2500 ;
	}
	
	*ddr_freq_khz = ( (1000*1000*1000)/(*ddr_period_ps) );	
}

/* Called from lib/board.c
 */
int dram_spd(void) {

	int 	node, chan, dimm;
	u8 	spd[NUM_SPD_BYTES];
	u32 	ddr_freq_khz, ddr_period_ps;
	__u64     clear_size ;

	volatile u32* mmio;

	/* Init memory structure    
	 */
	for (node=0; node<MAX_NODES; node++) {
		for (chan=0; chan<DDR_NUM_CHANNELS; chan++) {
			chan_mem_size[node][chan] = 0;
		}
	}

	for (node=0; node<num_nodes; node++) {

		for (chan=0; chan<DDR_NUM_CHANNELS; chan++) {

			ddr_cntrl[chan].num_ranks = 0;
			ddr_cntrl[chan].tck_min = 0;

			for (dimm=0; dimm<DDR_NUM_DIMMS; dimm++) {

				/* Populate spd_data[] with values 
				 * read in from the DIMM's EEPROM
				 */  
				if ( (dimm_to_spd_addr[node][chan][dimm].bus >= 0) &&
				     (read_ddr_spd(
					     dimm_to_spd_addr[node][chan][dimm].bus,
					     dimm_to_spd_addr[node][chan][dimm].spd_addr, 
					     0, 
					     NUM_SPD_BYTES, 
					     spd) ) == 0) {

					/* Verify SPD integrity and type    
					 */
					verify_spd(spd);
					/* Load SPD into params structure    
					 */
					param_load(&ddr_cntrl[chan],spd,dimm);
				}
			}

			if (ddr_cntrl[chan].num_ranks > 0) {
				chan_mem_size[node][chan] = 
					((ddr_cntrl[chan].node_density_mb * 
					 (ddr_cntrl[chan].mod_width/ddr_cntrl[chan].dev_width) * 
					  ddr_cntrl[chan].num_ranks)/8);
			} else {
				chan_mem_size[node][chan] = 0;
			}
			if (chan_mem_size[node][chan] != 0) {
				printf("DDR SPD: Node %d Channel %d Mem size = %d MB\n",
				       node, chan, chan_mem_size[node][chan]);
			}
		}


		find_lowest_freq(node,
				 &ddr_freq_khz,
				 &ddr_period_ps);

		set_dram_freq(node, ddr_freq_khz);

		printf("DDR SPD: Node %d DRAM frequency %d MHz\n",
				node, get_dram_freq(node)/1000);

		printf("DDR SPD: Node %d CPU frequency %d MHz\n",
				node, get_cpu_freq(node*32)/1000);

		for (chan=0; chan<DDR_NUM_CHANNELS; chan++) {

			if (chan_mem_size[node][chan] != 0) {

				/* Configure DDR Controller
				 */	
				spd_dbg("DDR SPD: Configuring DDR controller %d\n",chan);

				config(&ddr_cntrl[chan],node,chan,ddr_period_ps);

				/* Start DDR Init Sequence
				 */	
				spd_dbg("DDR SPD: Starting DDR controller %d\n",chan);
#ifndef XLP_SIM
				ddr_init_seq(node,chan);
#endif				
				if (ddr_cntrl[chan].registered) {
					write_cmr(&ddr_cntrl[chan],node,chan,ddr_period_ps) ;
					ddr_ucs_init_seq(dmc_io_mmio(node,chan) ) ;
				}
			}
		}
		

#ifndef XLP_SIM
		/*  Read/Write/Addr/Cmd Leveling Sequence
		 */		
	

		for (chan=0; chan<DDR_NUM_CHANNELS; chan++) {
			mmio = dmc_io_mmio(node, chan);
			if (chan_mem_size[node][chan] != 0) {
				ddr_read_level(node,chan,&ddr_cntrl[chan]) ;
				//ddr_print_linfo(node,chan,&ddr_cntrl[chan]) ;
				ddr_write_level(node,chan,&ddr_cntrl[chan]) ;
				//ddr_print_linfo(node,chan,&ddr_cntrl[chan]) ;
				ddr_adj_turnaround(node,chan,&ddr_cntrl[chan]) ; 
				ddr_addrcmd_align(node,chan,&ddr_cntrl[chan]) ;	
				//ddr_print_linfo(node,chan,&ddr_cntrl[chan]) ;
			}
		}

#endif	
		for (chan=0; chan<DDR_NUM_CHANNELS; chan++) {

			if (chan_mem_size[node][chan] != 0) {

				mmio = dmc_io_mmio(node, chan);
				
				spd_dbg("DDR SPD: Node %d Chan %d Starting Memory Clear\n",
					node, chan);
#ifndef XLP_HWEMU
				clear_size = ( (__u64) chan_mem_size[node][chan]) << 20 ;
#else
				clear_size =  1024*1024 ;
#endif
#ifndef XLP_SIM
				ddr_tge_memclear(mmio,
						 0,
						 clear_size) ;
#endif
				spd_dbg("DDR SPD: Node %d Chan %d Memory Clear Complete\n",
					node,chan);


#ifndef XLP_HWEMU
				/* Enable ECC if necessary */
				if (ddr_cntrl[chan].ecc_dev) {
					ddr_enable_ecc(mmio);
				}				
#endif
				
				ddr_enable_selfref(mmio) ;
				ddr_enable_pwrdown(mmio) ;
			}
		}
	}
	
	return 0;
}
