/*
 * Copyright 2003-2010 Netlogic Microsystems (“Netlogic”). All rights
 * reserved. Redistribution and use in source and binary forms, with 
 * or without modification, are permitted provided that the following 
 * conditions are met:
 *
 *	1. 	Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 *
 *	2. 	Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in
 *		the documentation and/or other materials provided with the
 *		distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * --------------------------------#NETL_2#--------------------------------
 */

#include <common.h>				
#include <asm/arch/cpu.h>

/* Low-level routines
 */
static inline __u8 oc_read_reg(int bus,int offset) {

	volatile __u32 *i2c_mmio = cpu_io_mmio(bus/2, (bus%2) ? I2C1:I2C0);
	return ((__u8)i2c_mmio[offset]);
}

static inline void oc_write_reg(int bus,int offset, __u8 value) {
	
	volatile __u32 *i2c_mmio = cpu_io_mmio(bus/2, (bus%2) ? I2C1:I2C0);
	i2c_mmio[offset] = value;
}

static inline int xaction_complete(int bus) {

	volatile int timeout = I2C_TIMEOUT;
	int retval = 0;

	while ((oc_read_reg(bus, OC_STATUS) & OC_STATUS_TIP) && timeout) {
		timeout--;
	}
	if (timeout == 0) {
		printf("Timed Out Waiting for TIP to Clear.\n");
		retval = -1;
	}
	return retval;
}

static inline int bus_idle(int bus) {

	volatile int timeout = I2C_TIMEOUT;
	int retval = 0;

	while ((oc_read_reg(bus, OC_STATUS) & OC_STATUS_BUSY) && timeout) {
		timeout--;
	}
	if (timeout == 0) {
		printf("Timed Out Waiting for Bus Busy to Clear.\n");
		retval = -1;
	}
	return retval;
}

static inline int check_for_ack(int bus) {
	
	if (oc_read_reg(bus, OC_STATUS) & OC_STATUS_NACK) {
		//printf("ACK not received from Slave Device.\n");
		return -1;
	}
	return 0;
}

/* ------------------------------------------
 * Sequence to Read byte(s) from an 
 * offset within the Slave Device: -
 * ------------------------------------------
 * -- Set up 'DATA' with 'Slave Address + W'
 * -- Trigger the START Condition
 * 	-> Wait for Xfer complete & Slave Ack
 * -- Set up 'DATA' with 'Slave Offset + W'
 * -- Set the 'WRITE' bit in COMMAND Reg
 * 	-> Wait for Xfer complete & Slave Ack
 * -- Set up 'DATA' with 'Slave Address + R'
 * -- Trigger the (Re-)START Condition
 * 	-> Wait for Xfer complete & Slave Ack
 * -- For 'n-1' out of 'n' bytes, 
 * 	-> Set 'READ' bit in COMMAND 
 *	-> Keep Reading 'DATA' Register
 * -- For the last (or only) byte
 * 	-> Set 'READ/STOP/NACK' bits in COMMAND 
 *	-> Read 'DATA' Register
 * ------------------------------------------
 */

int oc_i2c_rx(int bus, __u8 slave_addr, __u16 slave_offset, 
	      int len, __u8 *data) {

	/* Verify the bus is idle */
	if (oc_read_reg(bus, OC_STATUS) & OC_STATUS_BUSY) {
		printf("I2C Bus BUSY (Not Available), Aborting.\n");
		goto i2c_rx_error;
	}

	oc_write_reg(bus, OC_DATA, (slave_addr << 1) | WRITE_BIT);
	oc_write_reg(bus, OC_COMMAND, OC_COMMAND_START);
	if (xaction_complete(bus) < 0) {
		goto i2c_rx_error;
	}
	if (check_for_ack(bus) < 0) {
		goto i2c_rx_error;
	}

	/* Verify Arbitration is not Lost */
	if (oc_read_reg(bus, OC_STATUS) & OC_STATUS_AL) {
		printf("I2C Bus Arbitration Lost, Aborting.\n");
		goto i2c_rx_error;
	}

	oc_write_reg(bus, OC_DATA, slave_offset | WRITE_BIT);
	oc_write_reg(bus, OC_COMMAND, OC_COMMAND_WRITE);
	if (xaction_complete(bus) < 0) {
		goto i2c_rx_error;
	}
	if (check_for_ack(bus) < 0) {
		goto i2c_rx_error;
	}

	/* Address Phase Done, Data Phase begins
	 */
	oc_write_reg(bus, OC_DATA, (slave_addr << 1) | READ_BIT);
	oc_write_reg(bus, OC_COMMAND, OC_COMMAND_START);
	if (xaction_complete(bus) < 0) {
		goto i2c_rx_error;
	}
	if (check_for_ack(bus) < 0) {
		goto i2c_rx_error;
	}

	if (len > 1) {

		int bytenr = 0;

		for (bytenr = 0; bytenr < len-1; bytenr++) {
			oc_write_reg(bus, OC_COMMAND, OC_COMMAND_READ);
			if (xaction_complete(bus) < 0) {
				goto i2c_rx_error;
			}
			*data = oc_read_reg(bus, OC_DATA);
			data++;
		}
	}

	/* Last (or only) Byte: -
	 * 	Set RD, NACK, STOP Bits
	 */
	oc_write_reg(bus, OC_COMMAND, OC_COMMAND_STOP | OC_COMMAND_RDNACK);
	if (xaction_complete(bus) < 0) {
		goto i2c_rx_error;
	}

	*data = oc_read_reg(bus, OC_DATA);

	return bus_idle(bus);	

i2c_rx_error:
	/* Release Bus */
	oc_write_reg(bus, OC_COMMAND, OC_COMMAND_STOP);	
	bus_idle(bus);
	return -1;
}

/* Initialization for the Controller: -
 * -- Disable Controller/Interrupts
 * -- Program PreScaler
 * -- Enable Controller (Keep Intrs disabled)
 */
int i2c_init (void)
{
	int i, prescaler;

	/* Two I2C controllers per chip 
	 */ 
	for (i=0; i<num_nodes*2; i++) {

		/* Change to 45KHz (instead of 100KHz) 
		 * in order to meet setup and hold time
		 */
		__u8 ctrl_reg;

		prescaler = (I2C_CLKFREQ_KHZ/(5 * 45e3)) - 1;
		ctrl_reg = oc_read_reg(i, OC_CONTROL);

		oc_write_reg(i, OC_CONTROL, 
				ctrl_reg & ~(OC_CONTROL_EN | OC_CONTROL_IEN));

		oc_write_reg(i, OC_PRESCALER_LO, prescaler & 0xff);
		oc_write_reg(i, OC_PRESCALER_HI, prescaler >> 8);

		oc_write_reg(i, OC_CONTROL, 
				oc_read_reg(i, OC_CONTROL) | OC_CONTROL_EN);

		if (oc_read_reg(i, OC_CONTROL) & OC_CONTROL_EN) {
			printf("Initialized I2C%d Controller.\n", i);
		} else {
			return -1;
		}
	}

	return 0;
}

/* Write byte(s) to the I2C Slave Device
 */
int oc_i2c_tx(int bus,__u8 slave_addr, __u16 slave_offset, __u8 alen,
	      int len, __u8 *data) {
        int i ;
	
        /* Verify the bus is idle */
        if (oc_read_reg(bus,OC_STATUS) & OC_STATUS_BUSY) {
                printf("I2C Bus BUSY (Not Available), Aborting.\n");
                goto i2c_tx_error;
        }
	
        oc_write_reg(bus,OC_DATA, (slave_addr << 1) | WRITE_BIT);
        oc_write_reg(bus,OC_COMMAND, OC_COMMAND_START);
        if (xaction_complete(bus) < 0) {
		printf("TX Error 1.\n");

                goto i2c_tx_error;
        }
        if (check_for_ack(bus) < 0) {
		printf("TX Error 2.\n");
                goto i2c_tx_error;
        }
	
        /* Verify Arbitration is not Lost */
        if (oc_read_reg(bus,OC_STATUS) & OC_STATUS_AL) {
                printf("I2C Bus Arbitration Lost, Aborting.\n");
                goto i2c_tx_error;
        }
	
        for (i = 0; i<=alen; i++) {
                oc_write_reg(bus,OC_DATA, ( (slave_offset >> (i*8) ) & 0xff) | WRITE_BIT);
                oc_write_reg(bus,OC_COMMAND, OC_COMMAND_WRITE);
                if (xaction_complete(bus) < 0) {
			printf("TX Error 3.\n");
                        goto i2c_tx_error;
                }
                if (check_for_ack(bus) < 0) {
			printf("TX Error 4.\n");
                        goto i2c_tx_error;
                }
        }
	
        if (len > 1) {
                int bytenr = 0;
		
                for (bytenr = 0; bytenr < len-1; bytenr++) {
                        oc_write_reg(bus,OC_DATA, *data);
                        oc_write_reg(bus,OC_COMMAND, OC_COMMAND_WRITE);
                        if (xaction_complete(bus) < 0) {
				printf("TX Error 5.\n");
                                goto i2c_tx_error;
                        }
                        data++;
                }
        }
        /* Last (or only) Byte: -
         *      Set WR, STOP Bits
         */
        oc_write_reg(bus,OC_DATA, *data);
        oc_write_reg(bus,OC_COMMAND, OC_COMMAND_STOP | OC_COMMAND_WRITE);
        if (xaction_complete(bus) < 0) {
                goto i2c_tx_error;
        }
        if (check_for_ack(bus) < 0) {
                goto i2c_tx_error;
        }

        return bus_idle(bus);
	
i2c_tx_error:
        /* Release Bus */
        oc_write_reg(bus,OC_COMMAND, OC_COMMAND_STOP);
        bus_idle(bus);
        return -1;
}
