/*
 * Copyright 2003-2010 Netlogic Microsystems (“Netlogic”). All rights
 * reserved. Redistribution and use in source and binary forms, with 
 * or without modification, are permitted provided that the following 
 * conditions are met:
 *
 *	1. 	Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 *
 *	2. 	Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in
 *		the documentation and/or other materials provided with the
 *		distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * --------------------------------#NETL_2#--------------------------------
 */

/* This file has the DRAM programming with
 * added hooks for supporting multiple nodes
 * and channels across dual I2C busses. 
 */

#include <common.h>
#include <asm/arch/cpu.h>
#include "clock.h"
#include "ddr_spd.h"
#include "ddr_helper.h"

#define NUM_SPD_BYTES   	256
#define ALLOC_LIST_SIZE         8

struct mem_holes {
	int32_t hole_offset;
	int32_t hole_size;	
};

struct mblock {
	int32_t size ;
	int32_t ileave ;
} ;
struct mem_part {
	struct mblock nmem;
	struct mblock cmem[MAX_NODES][ALLOC_LIST_SIZE] ;
} ;

#define NODE_INTERLEAVE 1
#define CHAN_INTERLEAVE 1

/* 0      -   320 MB
 * 320MB  -   464  MB (144 MB Hole)
 * 464MB  -   3072 MB
 * 3072MB -   3584 MB (512 MB Hole)
 */
static const struct mem_holes dram_hole_map[] = 
{
	{ 320,      144 },
	{ 3072,     512 },
	{ 1048576,    0 }
};

static int setup_node_blocks(struct mem_part *npart)
{
	int node ;
	u32 nmem[MAX_NODES]; 
	int index = 0 ;
	int ways = 0 ;
	int chan ;

	for (index=0; index<ALLOC_LIST_SIZE; index++) {
		npart[index].nmem.size = 0;
	}

	for (node = 0;node < MAX_NODES;node++) {
		nmem[node] = 0;
		for (chan = 0;chan < DDR_NUM_CHANNELS; chan++) {
			nmem[node] += chan_mem_size[node][chan];
		}
		if (nmem[node] > 0) {
			ways |= (1 << node) ;
		}
	}
	index = 0;
	while ( (index != ALLOC_LIST_SIZE) &&
		(ways != 0) ) {

		for (node = 0;node < num_nodes ;node++) {
			if ( (nmem[node] != 0) &&
			     ( (npart[index].nmem.size > nmem[node]) ||
			       (npart[index].nmem.size == 0) ) ) {
				npart[index].nmem.size   = nmem[node] ;
				npart[index].nmem.ileave = node ;
			}
		}

		if (NODE_INTERLEAVE && (ways == 0xf) ) {
			npart[index].nmem.ileave = NBU_DDR_NODE_INTMODE_0_1_2_3 ;
			nmem[0] -= npart[index].nmem.size;
			nmem[1] -= npart[index].nmem.size;
			nmem[2] -= npart[index].nmem.size;
			nmem[3] -= npart[index].nmem.size;
			npart[index].nmem.size *= 4;
			
		} else if (NODE_INTERLEAVE && 
			   ( (ways & 0x3) == 0x3) && 
			   (npart[index].nmem.ileave < 2) ) {
			npart[index].nmem.ileave = NBU_DDR_NODE_INTMODE_0_1 ;
			nmem[0] -= npart[index].nmem.size;
			nmem[1] -= npart[index].nmem.size;
			npart[index].nmem.size *= 2;						
		} else if (NODE_INTERLEAVE && 
			   ( (ways & 0xc) == 0xc) && 
			   (npart[index].nmem.ileave >= 2) ) {
			npart[index].nmem.ileave = NBU_DDR_NODE_INTMODE_2_3 ;
			nmem[2] -= npart[index].nmem.size;
			nmem[3] -= npart[index].nmem.size;
			npart[index].nmem.size *= 2;			
		} else if (NODE_INTERLEAVE && 
			   ( (ways & 0x5) == 0x5) &&
			   ( (npart[index].nmem.ileave & 0x1) == 0) ) {
			npart[index].nmem.ileave = NBU_DDR_NODE_INTMODE_0_2 ;
			nmem[0] -= npart[index].nmem.size;
			nmem[2] -= npart[index].nmem.size;
			npart[index].nmem.size *= 2;
		} else {
			nmem[npart[index].nmem.ileave] -= npart[index].nmem.size;
		}

		for (node = 0;node < num_nodes ;node++) {
			if (nmem[node] == 0) {
				ways &= ~(1<<node) ;	
			}
		}
		index++ ;
	}
	if ( (index == ALLOC_LIST_SIZE) ||
	     (ways != 0) ) {
		printf("Node bar allocation faile to complete\n") ;
		return -1 ;
	} 
	return 0;
}

void setup_chan_blocks(u32 node,struct mem_part *npart)
{
	int ways ;
	int index = 0 ;
	int i ;
	int chan;
	int size ;

	for (i = 0 ; i < ALLOC_LIST_SIZE; i++) {
		npart->cmem[node][i].size = 0 ;	
	}

	if ( (npart->nmem.ileave == NBU_DDR_NODE_INTMODE_0_1_2_3) ||
	     ( (npart->nmem.ileave == NBU_DDR_NODE_INTMODE_0_1) &&
	       (node < 2) ) ||
	     ( (npart->nmem.ileave == NBU_DDR_NODE_INTMODE_2_3) &&
	       (node >= 2) ) ||
	     ( (npart->nmem.ileave == NBU_DDR_NODE_INTMODE_0_2) &&
	       ( (node & 0x1) == 0) ) ||
	     (node == npart->nmem.ileave) ) {
		
		ways = 0 ;
		for (chan = 0;chan < DDR_NUM_CHANNELS; chan++) {
			if (chan_mem_size[node][chan] > 0) {
				ways |= (1<<chan) ;
			}
		}
		size = npart->nmem.size ;
		while ( (index != ALLOC_LIST_SIZE) &&
			(ways != 0) && 
			(size > 0) ) {

			for (chan = 0;chan <  DDR_NUM_CHANNELS;chan++) {
				if ( (chan_mem_size[node][chan] != 0) &&
				     ( (npart->cmem[node][index].size > chan_mem_size[node][chan]) ||
				       (npart->cmem[node][index].size == 0) ) ) {
					npart->cmem[node][index].size = chan_mem_size[node][chan] ;
					npart->cmem[node][index].ileave = chan ;
				}
			}
			if (size < npart->cmem[node][index].size) {
				npart->cmem[node][index].size = size;
			}
			if (CHAN_INTERLEAVE && (ways == 0xf) ) {

				npart->cmem[node][index].ileave = NBU_DDR_CHAN_INTMODE_0_1_2_3 ;
				if (size >= (npart->cmem[node][index].size*4) ) {
					chan_mem_size[node][0] -= npart->cmem[node][index].size;
					chan_mem_size[node][1] -= npart->cmem[node][index].size;
					chan_mem_size[node][2] -= npart->cmem[node][index].size;
					chan_mem_size[node][3] -= npart->cmem[node][index].size;
					npart->cmem[node][index].size = (npart->cmem[node][index].size) * 4;	
				} else {
					chan_mem_size[node][0] -= (size/4);
					chan_mem_size[node][1] -= (size/4);
					chan_mem_size[node][2] -= (size/4);
					chan_mem_size[node][3] -= (size/4);
					npart->cmem[node][index].size = (size/4) * 4;	
				}			       
			} else if (CHAN_INTERLEAVE && 
				   ( (ways & 0x3) == 0x3)   && 
				   (npart->cmem[node][index].ileave < 2) ) {

				npart->cmem[node][index].ileave  = NBU_DDR_CHAN_INTMODE_0_1 ;
				if (size >= (npart->cmem[node][index].size*2) ) {
					chan_mem_size[node][0] -= npart->cmem[node][index].size;
					chan_mem_size[node][1] -= npart->cmem[node][index].size;
					npart->cmem[node][index].size = (npart->cmem[node][index].size) * 2;	
				} else {
					chan_mem_size[node][0] -= (size/2);
					chan_mem_size[node][1] -= (size/2);
					npart->cmem[node][index].size = (size/2) * 2;	
				}			       
			} else if (CHAN_INTERLEAVE && 
				   ( (ways & 0xc) == 0xc)   && 
				   (npart->cmem[node][index].ileave >= 2) ) {

				npart->cmem[node][index].ileave   = NBU_DDR_CHAN_INTMODE_2_3 ;
				if (size >= (npart->cmem[node][index].size*2) ) {
					chan_mem_size[node][2] -= npart->cmem[node][index].size;
					chan_mem_size[node][3] -= npart->cmem[node][index].size;
					npart->cmem[node][index].size = (npart->cmem[node][index].size) * 2;	
				} else {
					chan_mem_size[node][2] -= (size/2);
					chan_mem_size[node][3] -= (size/2);
					npart->cmem[node][index].size = (size/2) * 2;	
				}
			} else if (CHAN_INTERLEAVE && 
				   ( (ways & 0x5) == 0x5) && 
				   ( (npart->cmem[node][index].ileave & 0x1) == 0) ) {
				npart->cmem[node][index].ileave   = NBU_DDR_CHAN_INTMODE_0_2 ;
				if (size >= (npart->cmem[node][index].size*2) ) {
					chan_mem_size[node][0] -= npart->cmem[node][index].size;
					chan_mem_size[node][2] -= npart->cmem[node][index].size;
					npart->cmem[node][index].size = (npart->cmem[node][index].size) * 2;	
				} else {
					chan_mem_size[node][0] -= (size/2);
					chan_mem_size[node][2] -= (size/2);
					npart->cmem[node][index].size = (size/2) * 2;	
				}
			} else {
				chan_mem_size[node][npart->cmem[node][index].ileave] -= npart->cmem[node][index].size;
			}

			size -= (npart->cmem[node][index].size) ;
			
			for (chan = 0;chan < DDR_NUM_CHANNELS ;chan++) {
				if (chan_mem_size[node][chan] == 0) {
					ways &= ~(1<<chan) ;	
				}
			}
			index++ ;
		}
		
		if ( (index == ALLOC_LIST_SIZE) ||
		     (ways != 0) )  {
			printf("Chan bar allocation faile to complete %08x\n",size) ;
			return  ;
		} 
	} else {
		npart->cmem[node][0].size   = npart->nmem.size ;
		npart->cmem[node][0].ileave = 0 ;
	}	
}
void setup_mem_part(struct mem_part *npart) 
{
	int index = 0;
	int node ;

	setup_node_blocks(npart) ;

	while ( (index != ALLOC_LIST_SIZE) &&
		(npart[index].nmem.size != 0) ) {
		for (node = 0;node < num_nodes ;node++) {
			setup_chan_blocks(node,&npart[index]) ;
			
		}
		index++ ;
	}
}

int setup_bridge_bars(void) {

	int 	node, i=0,bar[MAX_NODES];
	int32_t mem_offset[MAX_NODES], hole_total[MAX_NODES];
	int32_t hole_index[MAX_NODES], hole_size;
	int32_t region_size=0;
	int32_t nindex,cindex;
	
	volatile u32* mmio;
	volatile u32* gbu_mmio;

	struct mem_part npart[ALLOC_LIST_SIZE]; 

	for (node=0; node<num_nodes; node++) {
		mmio = (volatile u32*) cpu_io_mmio(node, BRIDGE);

		/* Invalidate all DRAM BARs 
		 */
		for (i=0; i<NUM_DRAM_BARS; i++) {
			mmio[NBU_DRAM_BAR_LIMIT_0 + i] = 0x0;
			mmio[NBU_DRAM_BAR_0 + i]       = 0xfffff000;
		}
		bar[node] = 0;
		mem_offset[node] = 0;
		hole_total[node] = 0;
		hole_index[node] = 0;
	}

	for (node=0; node<num_nodes; node++) {
		mmio     = cpu_io_mmio(node,BRIDGE);
		gbu_mmio = cpu_io_mmio(node,GBU);
	
		/*
		 * Disable DMC read credit limiting.
		 */
		mmio[NBU_MODE_REG] = mmio[NBU_MODE_REG] | NBU_DMC_RDCRD_DISABLE ;

		/* Re-map flash to 0x16000000 - 0x17ffffff (32M)
		 */
		mmio[NBU_FLASH_BAR_LIMIT_0]         = 0x0;
		mmio[NBU_FLASH_BAR_0]               = 0x160000;
		mmio[NBU_FLASH_BAR_LIMIT_0]         = 0x17f000;
		if (node == 0) {
			gbu_mmio[GBU_CS0_BASEADDRESS_REG]   = 0x160000;
			gbu_mmio[GBU_CS0_BASELIMIT_REG]     = 0x17ffff;
		} 
	}
	setup_mem_part(npart);

	nindex = 0 ;
	while (npart[nindex].nmem.size != 0) {
		for (node = 0; node<num_nodes;node++) {
			cindex = 0;				
			while (npart[nindex].cmem[node][cindex].size != 0) {
				if (bar[node] == 8) {
					printf("Error! Unable to allocate Bridge BARs for node %d\n",node);	
					return -1;
				}
				
				if (((mem_offset[node] + npart[nindex].cmem[node][cindex].size) 
				     - dram_hole_map[hole_index[node]].hole_offset) > 0) {
					region_size = dram_hole_map[hole_index[node]].hole_offset - mem_offset[node];
					hole_size   = dram_hole_map[hole_index[node]].hole_size;
					hole_index[node]++;
				} else {
					region_size = npart[nindex].cmem[node][cindex].size;
					hole_size   = 0;
				}
							
				mmio = (volatile u32*) cpu_io_mmio(node, BRIDGE);
					
				mmio[NBU_DRAM_BAR_0 + bar[node]]         = (mem_offset[node] << 12);
				mmio[NBU_DRAM_BAR_LIMIT_0 + bar[node]]   = (mem_offset[node] + region_size - 1) << 12;
				mmio[NBU_DRAM_BAR_XLAT_0 + bar[node]]    = 
					 ( (hole_total[node] << 12) | 
					   (npart[nindex].cmem[node][cindex].ileave << 1) | 
					   0x1);
				
				mmio[NBU_DRAM_BAR_NODEADDR_XLAT_0 + bar[node]]  = 
					( (npart[nindex].nmem.ileave  << 1) 
					  | 0x1);
				
				printf("NBU%d DRAM BAR%d base: %08x limit: %08x xlate: %08x node: %08x (% 5d MB -> % 5d MB, size: % 5d MB)\n",
				       node,
				       bar[node],
				       mmio[NBU_DRAM_BAR_0 + bar[node]],
				       mmio[NBU_DRAM_BAR_LIMIT_0 + bar[node]],
				       mmio[NBU_DRAM_BAR_XLAT_0 + bar[node]],
				       mmio[NBU_DRAM_BAR_NODEADDR_XLAT_0 + bar[node]],
				       mem_offset[node], mem_offset[node] + region_size, region_size);
				
				npart[nindex].cmem[node][cindex].size -= region_size ;
								
				if (npart[nindex].cmem[node][cindex].size == 0)
					cindex++;
				
				hole_total[node] += hole_size;			
				mem_offset[node]  = mem_offset[node] + region_size + hole_size;
				
				bar[node]++;
			} 
		}
		nindex++ ;
	}
		return 0;
}

/* Called from lib/board.c
 */
int dram_init(void) {

	/* Workaround for colliding I2C address on
	 * EVP1 eval board.
	 */
#ifndef XLP_SIM
#ifdef CONFIG_SYS_I2C_MUX
	int node ;
	u8 data=0 ;
	for (node=0; node<num_nodes; node++) {		
		/* Disable on board I2C Mux*/
		oc_i2c_tx(1 + (node*2),
			  0x70,
			  0,
			  0,
			  1,
			  &data) ;
		oc_i2c_tx(0 + (node*2),
			  0x70,
			  0,
			  0,
			  1,
			  &data) ;
		
		oc_i2c_tx(0 + (node*2),
			  0x71,
			  0,
			  0,
			  1,
			  &data) ;
	}
#endif
#endif

	if (dram_spd() != 0) {
		return -1;
	}

	/* Configure Bridge Dram Bars
	 */	
	if (setup_bridge_bars() != 0) { 
		return -1; 
	}
	return 0;
}
